/*
 * rootapps graphics main management functions
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/time.h>
#include "options.h"
#include "rootapp.h"


#define HASAPP if (! app) {fprintf(stderr, "Application not initialized !\n"); exit(1);}


static int        screen;
static GC         gc;



unsigned long get_color(const char *color_name) {
    Colormap      colormap;
    XColor        color;
    int           screen = DefaultScreen(display);
    unsigned long res    = WhitePixel(display, screen);

    colormap = DefaultColormap(display, DefaultScreen(display));

    if (color_name) {
        if (! XParseColor(display, colormap, color_name, &color))
            fprintf(stderr, "Cannot find color \"%s\" !\n", color_name);
        else if (! XAllocColor(display, colormap, &color))
            fprintf(stderr, "Cannot allocate color \"%s\" !\n", color_name);
        else {
            res = color.pixel;
        }
    }

    return res;
}


void set_color(const char *color_name) {
    XSetForeground(display, gc, get_color(color_name));
}


void load_font(const char *font_name) {
    Font font;

    if (! font_name) return;
    font = XLoadFont(display, font_name);
    XSetFont(display, gc, font);
}


int font_height() {
    XFontStruct *fontstruct;
    GContext     ctxt;
    int          res = 0;

    ctxt = XGContextFromGC(gc);
    if ((fontstruct = XQueryFont(display, ctxt)) == NULL) {
        fprintf(stderr, "Cannot get font infos !\n");
        exit(1);
    }
    res = fontstruct->ascent + fontstruct->descent;
    XFreeFontInfo(NULL, fontstruct, 1);
    return res;
}


int font_ascent() {
    XFontStruct *fontstruct;
    GContext     ctxt;
    int          res = 0;

    ctxt = XGContextFromGC(gc);
    if ((fontstruct = XQueryFont(display, ctxt)) == NULL) {
        fprintf(stderr, "Cannot get font infos !\n");
        exit(1);
    }
    res = fontstruct->ascent;
    XFreeFontInfo(NULL, fontstruct, 1);
    return res;
}


int font_descent() {
    XFontStruct *fontstruct;
    GContext     ctxt;
    int          res = 0;

    ctxt = XGContextFromGC(gc);
    if ((fontstruct = XQueryFont(display, ctxt)) == NULL) {
        fprintf(stderr, "Cannot get font infos !\n");
        exit(1);
    }
    res = fontstruct->descent;
    XFreeFontInfo(NULL, fontstruct, 1);
    return res;
}


int text_width(const char *text) {
    XFontStruct *fontstruct;
    GContext     ctxt;
    int          res = 0;

    ctxt = XGContextFromGC(gc);
    if ((fontstruct = XQueryFont(display, ctxt)) == NULL) {
        fprintf(stderr, "Cannot get font infos !\n");
        exit(1);
    }
    res = XTextWidth(fontstruct, text, strlen(text)) + 2;
    XFreeFontInfo(NULL, fontstruct, 1);
    return res;
}


Pixmap copy_pixmap(RootApp *app, Pixmap source) {
    Pixmap   result = None;
    Window   parent;
    unsigned px, py, pw, ph, pb, depth;

    HASAPP;
    if (source == None) {
        fprintf(stderr, "Trying to copy a nonexistant pixmap !\n");
        return None;
    }
    XGetGeometry(display, source, &parent, &px, &py, &pw, &ph, &pb, &depth);
    result = XCreatePixmap(display, app->window, pw, ph, depth);
    XCopyArea(display, source, result, gc, 0, 0, pw, ph, 0, 0);

    return result;
}


Pixmap root_pixmap(const char *desc) {
    Pixmap res = None;
    Atom   id  = None;

    id = XInternAtom(display, desc, True);
    if (id != None) {
        Window         rootwin;
        Atom           act_type;
        int            act_format;
        unsigned long  nitems, bytes_after;
        unsigned char *prop = NULL;

        rootwin = DefaultRootWindow(display);
        if (XGetWindowProperty(display, rootwin, id, 0, 1, False,
                               XA_PIXMAP, &act_type, &act_format,
                               &nitems, &bytes_after, &prop) == Success) {
            if (prop) {
                res = *((Pixmap*) prop);
            }
        }
    }

    return res;
}


Pixmap get_root_pixmap(RootApp *app) {
    Pixmap res;
    
    HASAPP;
    res = root_pixmap("_XROOTPMAP_ID");
    if (res == None) res = root_pixmap("_XSETROOT_ID");

    if (res != None) {
        Pixmap z, winpixmap;
        Window parent;
        int    build = 0;
        unsigned int px, py, pw, ph, pb, pd, sw, sh, dx, dy;

        sw = DisplayWidth(display, screen);
        sh = DisplayHeight(display, screen);
        dx = 0;
        dy = 0;
        XGetGeometry(display, res, &parent, &px, &py, &pw, &ph, &pb, &pd);
        if ((pw < sw * 2) || (ph < sh * 2)) {
            z = XCreatePixmap(display, app->window, sw * 2, sh * 2, pd);
            while (dy < sh * 2) {
                while (dx < sw * 2) {
                    int w, h;
                    w = sw*2 - dx > pw ? pw : sw*2 - dx;
                    h = sh*2 - dy > ph ? ph : sh*2 - dy;
                    XCopyArea(display, res, z, gc, 0, 0, w, h, dx, dy);
                    dx += pw;
                }
                dy += ph;
                dx = 0;
            }
            res = z;
            build = 1;
        }
        winpixmap = XCreatePixmap(display, app->window,
                app->wsize.width, app->wsize.height, pd);
        XCopyArea(display, res, winpixmap, gc,
                app->wsize.x < 0 ? app->wsize.x + sw : app->wsize.x,
                app->wsize.y < 0 ? app->wsize.y + sh : app->wsize.y,
                app->wsize.width, app->wsize.height, 0, 0);
        if (build) XFreePixmap(display, res);
        res = winpixmap;
    } else {
        fprintf(stderr, "Could not get root pixmap !\n");
    }
    return res;
}


void clean_window(RootApp *app) {
    HASAPP;
    if (app->rootpixmap) {
        XCopyArea(display, app->rootpixmap, app->window, gc,
                0, 0, app->wsize.width, app->wsize.height,
                app->wsize.x, app->wsize.y);
        XFlush(display);
    }
}


void clear_bg(RootApp *app) {
    HASAPP;
    if ((app->rootpixmap) && (app->bgpixmap)) {
        XCopyArea(display, app->rootpixmap, app->bgpixmap, gc,
                0, 0, app->wsize.width, app->wsize.height, 0, 0);
    }
}


void display_app(RootApp *app) {
    HASAPP;
    if (app->bgpixmap) {
        XCopyArea(display, app->bgpixmap, app->window, gc,
                0, 0, app->wsize.width, app->wsize.height,
                app->wsize.x, app->wsize.y);
        XFlush(display);
    }
}


void get_geom(const char *g, XSizeHints *s, const char *text) {
    int nx, ny, nw, nh, status;
    int sw, sh;

    if (! g) return;
    sw = DisplayWidth(display, screen);
    sh = DisplayHeight(display, screen);

    nx = 0;
    ny = 0;
    status = XParseGeometry(g, &nx, &ny, &nw, &nh);
    if (!(status & WidthValue) && text) nw = text_width(text);
    if (!(status & HeightValue)) nh = font_height();
    if (status & XNegative) nx += sw - nw;
    if (status & YNegative) ny += sh - nh;

    s->x = nx;
    s->y = ny;
    s->width = nw;
    s->height = nh;
}


void root_print(RootApp *app, char *data, int x, int y) {
    HASAPP;
    XDrawString(display, app->bgpixmap, gc,
            x, y + font_ascent(),
            data, strlen(data));
}


RootApp *create_app(int mask,
        const char *geometry, const char *fontname, const char *text) {
    RootApp *app = xmalloc(sizeof(RootApp));

    if ((display = XOpenDisplay(NULL)) == NULL) {
        fprintf(stderr, "Cannot open display !\n");
        exit(1);
    }

    screen = DefaultScreen(display);
    app->window = RootWindow(display, screen);
    if (! gc) gc = XCreateGC(display, app->window, 0, NULL);
    XSelectInput(display, app->window, mask);
    XSetGraphicsExposures(display, gc, False);
    load_font(fontname);
    get_geom(geometry, &(app->wsize), text);
    app->rootpixmap = get_root_pixmap(app);
    app->bgpixmap = copy_pixmap(app, app->rootpixmap);

    XMapWindow(display, app->window);

    return app;
}


void close_app(RootApp *app) {
    HASAPP;
    clean_window(app);
    XFreePixmap(display, app->rootpixmap);
    XFreePixmap(display, app->bgpixmap);
    XFreeGC(display, gc);
    XDestroyWindow(display, app->window);
    XCloseDisplay(display);
    free(app);
}


int nextevent_or_timeout(XEvent *event, unsigned long miliseconds) {
    struct timeval timeout;
    fd_set rset;

    if (XPending(display) || (miliseconds == 0)) {
        XNextEvent(display, event);
        return True;
    }

    timeout.tv_sec = miliseconds / 1000;
    timeout.tv_usec = (miliseconds % 1000) * 1000;

    FD_ZERO(&rset);
    FD_SET(ConnectionNumber(display), &rset);
    if (select(ConnectionNumber(display)+1, &rset, NULL, NULL, &timeout) > 0) {
        XNextEvent(display, event);
        return True;
    }

    return False;
}

